function z = glad_sharpen(y, h, C, LQ, ksize, tau1, tau2, beta)
% GLAD for grayscale images.
%
% [RETURNS]
% z     : the restored image
%
% [PARAMETERS]
% y     : the input image
% h     : the global smoothing parameter
% C     : the covariance matrices containing local orientation information
% LQ    : the local metric Q matrix.
% ksize : the size of the kernel (ksize x ksize, and "ksize" must be
%         an odd number)
% tau1, tau2, beta: parameters for setting sharpness constant kappa

r = 1;

% Get the original image size
[N, M] = size(y);

% Initialize the return parameters
z = zeros(N*r, M*r);

% Pixel sampling positions
radius = (ksize - 1) / 2;
[x2, x1] = meshgrid(-radius-(r-1)/r : 1/r : radius, -radius-(r-1)/r : 1/r : radius);

% pre-culculation for covariance matrices
C11 = zeros(N, M);
C12 = zeros(N, M);
C22 = zeros(N, M);
sq_detC = zeros(N, M);

for n = 1 : N
    for m = 1 : M
        C11(n,m) = C(1,1,n,m);
        C12(n,m) = C(1,2,n,m);
        C22(n,m) = C(2,2,n,m);
        sq_detC(n,m) = sqrt(det(C(:,:,n,m)));
    end
end

% Mirroring
y = EdgeMirror(y, [radius, radius]);
C11 = EdgeMirror(C11, [radius, radius]);
C12 = EdgeMirror(C12, [radius, radius]);
C22 = EdgeMirror(C22, [radius, radius]);
sq_detC = EdgeMirror(sq_detC, [radius, radius]);

% Laplacian Kernel
h2 = 0.4; %h2 = 0.3;
[lx2 lx1] = meshgrid(-2:1:2, -2:1:2);
ltt = (lx1.^2+lx2.^2) ./ (2*h2^2);
L = -(1-ltt).*exp(-ltt)*8;
L(3,3) = 0;
L(3,3) = -sum(L(:));

% Estimate an image and its first gradients with pixel-by-pixel
for i = 1 : r
    for j = 1 : r
        xx1 = downsample2(x1(r-i+1:end, r-j+1:end), r);
        xx2 = downsample2(x2(r-i+1:end, r-j+1:end), r);
                
        for n = 1 : N
            nn = (n - 1) * r + i;
            for m = 1 : M
                mm = (m - 1) * r + j;
                
                % Neighboring samples to be taken account into the estimation
                yp = y(n:n+ksize-1, m:m+ksize-1);
                
                % compute the weight matrix
                tt = xx1 .* (C11(n:n+ksize-1, m:m+ksize-1) .* xx1...
                        + C12(n:n+ksize-1, m:m+ksize-1) .* xx2)...
                        + xx2 .* (C12(n:n+ksize-1, m:m+ksize-1) .* xx1...
                        + C22(n:n+ksize-1, m:m+ksize-1) .* xx2);
                W = exp(-(0.5/h^2) * tt) .* sq_detC(n:n+ksize-1, m:m+ksize-1)./(h^2);
                
                % Equivalent kernel
                E0 = W./sum(W(:));
                
                if LQ(n,m)<=tau1
                    kappa = beta;
                elseif LQ(n,m)>=tau2
                    kappa = 0;
                else
                    kappa = beta*(LQ(n,m)-tau2)/(tau1-tau2);
                end
                E0L = conv2(E0, L, 'same');
                E2 = E0 - E0L*kappa;
                E2 = E2./abs(sum(E2(:)));

                % Estimate the pixel values at (nn,mm)
                z(nn,mm)   = E2(:)' * yp(:);
            end
        end
    end
end